@@ -10,6 +10,8 @@ use crate::ui::{CsvTable, CsvTableState, FilterColumnsState, FinderState};
 use crate::view;
 
 use anyhow::ensure;
+#[cfg(feature = "clipboard")]
+use arboard::Clipboard;
 use ratatui::backend::Backend;
 use ratatui::{Frame, Terminal};
 
@@ -93,6 +95,38 @@ fn get_cols_offset_to_fill_frame_width(
     }
 }
 
+#[derive(Default)]
+pub struct LineWrapState {
+    pub enable_line_wrap: bool,
+    pub is_word_wrap: bool,
+}
+
+impl LineWrapState {
+    pub fn toggle(&mut self, is_word_wrap: bool) {
+        // Just switch between line wrap and word wrap if already enabled
+        if self.enable_line_wrap && self.is_word_wrap != is_word_wrap {
+            self.is_word_wrap = is_word_wrap;
+            return;
+        }
+
+        // Toggle enabling or disabling wrapping
+        self.enable_line_wrap = !self.enable_line_wrap;
+        self.is_word_wrap = is_word_wrap
+    }
+
+    pub fn transient_message(&self) -> String {
+        if self.enable_line_wrap {
+            if self.is_word_wrap {
+                "Word wrap enabled".to_string()
+            } else {
+                "Line wrap enabled".to_string()
+            }
+        } else {
+            "Line wrap disabled".to_string()
+        }
+    }
+}
+
 pub struct App {
     input_handler: InputHandler,
     num_rows_not_visible: u16,
@@ -108,6 +142,9 @@ pub struct App {
     ignore_case: bool,
     help_page_state: help::HelpPageState,
     sorter: Option<Arc<sort::Sorter>>,
+    line_wrap_state: LineWrapState,
+    #[cfg(feature = "clipboard")]
+    clipboard: Result<Clipboard>,
 }
 
 impl App {
@@ -166,6 +203,12 @@ impl App {
         let transient_message: Option<String> = None;
         let help_page_state = help::HelpPageState::new();
 
+        #[cfg(feature = "clipboard")]
+        let clipboard = match Clipboard::new() {
+            Ok(clipboard) => Ok(clipboard),
+            Err(e) => Err(anyhow::anyhow!(e)),
+        };
+
         let mut app = App {
             input_handler,
             num_rows_not_visible,
@@ -181,6 +224,9 @@ impl App {
             ignore_case,
             help_page_state,
             sorter: None,
+            line_wrap_state: LineWrapState::default(),
+            #[cfg(feature = "clipboard")]
+            clipboard,
         };
 
         if let Some(pat) = &columns_regex {
@@ -352,6 +398,19 @@ impl App {
             Control::Find(s) | Control::Filter(s) => {
                 self.handle_find_or_filter(s, matches!(control, Control::Filter(_)));
             }
+            Control::FindLikeCell | Control::FilterLikeCell => {
+                if let Some(value) = self.rows_view.get_cell_value_from_selection() {
+                    self.handle_find_or_filter(
+                        value.as_str(),
+                        matches!(control, Control::FilterLikeCell),
+                    );
+                } else {
+                    self.transient_message.replace(
+                        "Select a cell first before finding (#) or filtering (@) rows like it"
+                            .to_string(),
+                    );
+                }
+            }
             Control::FilterColumns(pat) => {
                 self.set_columns_filter(pat);
             }
@@ -367,17 +426,13 @@ impl App {
             Control::ToggleSelectionType => {
                 self.rows_view.selection.toggle_selection_type();
             }
-            Control::ToggleLineWrap => {
+            Control::ToggleLineWrap(word_wrap) => {
                 self.csv_table_state.reset_buffer();
-                if self.csv_table_state.enable_line_wrap {
-                    self.csv_table_state.enable_line_wrap = false;
-                    self.transient_message
-                        .replace("Line wrap disabled".to_string());
-                } else {
-                    self.csv_table_state.enable_line_wrap = true;
-                    self.transient_message
-                        .replace("Line wrap enabled".to_string());
-                }
+                self.line_wrap_state.toggle(*word_wrap);
+                self.csv_table_state.enable_line_wrap = self.line_wrap_state.enable_line_wrap;
+                self.csv_table_state.is_word_wrap = self.line_wrap_state.is_word_wrap;
+                self.transient_message
+                    .replace(self.line_wrap_state.transient_message());
             }
             Control::ToggleSort => {
                 if let Some(selected_column_index) = self.get_global_selected_column_index() {
@@ -392,10 +447,7 @@ impl App {
                     if should_create_new_sorter {
                         let column_name = self
                             .rows_view
-                            .raw_headers()
-                            .get(selected_column_index as usize)
-                            .cloned()
-                            .unwrap_or_default();
+                            .get_column_name_from_global_index(selected_column_index as usize);
                         let _sorter = sort::Sorter::new(
                             self.shared_config.clone(),
                             selected_column_index as usize,
@@ -414,6 +466,19 @@ impl App {
             Control::DecreaseWidth => {
                 self.adjust_column_width(-4);
             }
+            #[cfg(feature = "clipboard")]
+            Control::CopySelection => {
+                if let Some(selected) = self.rows_view.get_cell_value_from_selection() {
+                    match self.clipboard.as_mut().map(|c| c.set_text(&selected)) {
+                        Ok(_) => self
+                            .transient_message
+                            .replace(format!("Copied {} to clipboard", selected.as_str())),
+                        Err(e) => self
+                            .transient_message
+                            .replace(format!("Failed to copy to clipboard: {e}")),
+                    };
+                }
+            }
             Control::Reset => {
                 self.csv_table_state.column_width_overrides.reset();
                 self.reset_filter();
@@ -460,7 +525,18 @@ impl App {
             if should_create_new_finder {
                 let target = self.finder.as_ref().unwrap().target();
                 let sorter = self.sorter.clone();
-                self.create_finder(target, self.rows_view.is_filter(), sorter);
+                if let Some(finder) = &self.finder {
+                    // Inherit previous finder's column index if any, instead of using the current
+                    // selected column intended for sorter
+                    self.create_finder_with_column_index(
+                        target,
+                        self.rows_view.is_filter(),
+                        finder.column_index(),
+                        sorter,
+                    );
+                } else {
+                    self.create_finder(target, self.rows_view.is_filter(), sorter);
+                }
             }
         }
 
@@ -539,7 +615,23 @@ impl App {
     }
 
     fn create_finder(&mut self, target: Regex, is_filter: bool, sorter: Option<Arc<sort::Sorter>>) {
-        let _finder = find::Finder::new(self.shared_config.clone(), target, sorter).unwrap();
+        self.create_finder_with_column_index(
+            target,
+            is_filter,
+            self.get_global_selected_column_index().map(|x| x as usize),
+            sorter,
+        );
+    }
+
+    fn create_finder_with_column_index(
+        &mut self,
+        target: Regex,
+        is_filter: bool,
+        column_index: Option<usize>,
+        sorter: Option<Arc<sort::Sorter>>,
+    ) {
+        let _finder =
+            find::Finder::new(self.shared_config.clone(), target, column_index, sorter).unwrap();
         self.finder = Some(_finder);
         if is_filter {
             self.rows_view.set_rows_from(0).unwrap();
@@ -1086,7 +1178,7 @@ mod tests {
         let lines = to_lines(&actual_buffer);
         assert_eq!(lines, expected);
 
-        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap);
+        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap(false));
         let expected = vec![
             "──────────────────────────────────────────────────",
             "      a    b              c                       ",
@@ -1122,6 +1214,236 @@ mod tests {
         let actual_buffer = terminal.backend().buffer().clone();
         let lines = to_lines(&actual_buffer);
         assert_eq!(lines, expected);
+
+        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap(true));
+        let expected = vec![
+            "──────────────────────────────────────────────────",
+            "      a    b              c                       ",
+            "───┬─────────────────────────────────────┬────────",
+            "1  │  1    this is a      12345          │        ",
+            "   │       very long                     │        ",
+            "   │       text that                     │        ",
+            "   │       surely                        │        ",
+            "   │       will not                      │        ",
+            "   │       fit in                        │        ",
+            "   │       your small                    │        ",
+            "   │       screen                        │        ",
+            "2  │  2    this           678910         │        ",
+            "   │       is                            │        ",
+            "   │       an                            │        ",
+            "   │       even                          │        ",
+            "   │       longer                        │        ",
+            "   │       text                          │        ",
+            "   │       that                          │        ",
+            "   │       surely                        │        ",
+            "   │       will                          │        ",
+            "   │       not                           │        ",
+            "   │       fit                           │        ",
+            "   │       in                            │        ",
+            "   │       your                          │        ",
+            "   │       small                         │        ",
+            "   │       screen                        │        ",
+            "3  │  3    normal         123,456,789    │        ",
+            "   │       text now                      │        ",
+            "───┴─────────────────────────────────────┴────────",
+            "Word wrap enabled                                 ",
+        ];
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        assert_eq!(lines, expected);
+    }
+
+    #[test]
+    fn test_multiple_newlines() {
+        let mut app = AppBuilder::new("tests/data/multiple_newlines.csv")
+            .build()
+            .unwrap();
+        thread::sleep(time::Duration::from_millis(100));
+
+        let backend = TestBackend::new(50, 45);
+        let mut terminal = Terminal::new(backend).unwrap();
+
+        step_and_draw(&mut app, &mut terminal, Control::Nothing);
+        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap(true));
+        let expected = vec![
+            "──────────────────────────────────────────────────",
+            "      a    b              c                       ",
+            "───┬─────────────────────────────────────┬────────",
+            "1  │  1    this is a      12345          │        ",
+            "   │       very long                     │        ",
+            "   │       text that                     │        ",
+            "   │       surely                        │        ",
+            "   │       will not                      │        ",
+            "   │       fit in                        │        ",
+            "   │       your small                    │        ",
+            "   │       screen                        │        ",
+            "2  │  2    this           678910         │        ",
+            "   │                                     │        ",
+            "   │       is                            │        ",
+            "   │                                     │        ",
+            "   │       an                            │        ",
+            "   │                                     │        ",
+            "   │       even                          │        ",
+            "   │                                     │        ",
+            "   │       longer                        │        ",
+            "   │                                     │        ",
+            "   │       text                          │        ",
+            "   │                                     │        ",
+            "   │       that                          │        ",
+            "   │                                     │        ",
+            "   │       surely                        │        ",
+            "   │                                     │        ",
+            "   │       will                          │        ",
+            "   │                                     │        ",
+            "   │       not                           │        ",
+            "   │                                     │        ",
+            "   │       fit                           │        ",
+            "   │                                     │        ",
+            "   │       in                            │        ",
+            "   │                                     │        ",
+            "   │       your                          │        ",
+            "   │                                     │        ",
+            "   │       small                         │        ",
+            "   │                                     │        ",
+            "   │       screen                        │        ",
+            "3  │  3    normal         123,456,789    │        ",
+            "   │       text now                      │        ",
+            "   │                                     │        ",
+            "───┴─────────────────────────────────────┴────────",
+            "Word wrap enabled                                 ",
+        ];
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        assert_eq!(lines, expected);
+    }
+
+    #[test]
+    fn test_carriage_returns() {
+        let mut app = AppBuilder::new("tests/data/multi_lines_carriage_return.csv")
+            .build()
+            .unwrap();
+        thread::sleep(time::Duration::from_millis(100));
+
+        let backend = TestBackend::new(50, 45);
+        let mut terminal = Terminal::new(backend).unwrap();
+
+        step_and_draw(&mut app, &mut terminal, Control::Nothing);
+        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap(true));
+        let expected = vec![
+            "──────────────────────────────────────────────────",
+            "      a    b              c                       ",
+            "───┬─────────────────────────────────────┬────────",
+            "1  │  1    this is a      12345          │        ",
+            "   │       very long                     │        ",
+            "   │       text that                     │        ",
+            "   │       surely                        │        ",
+            "   │       will not                      │        ",
+            "   │       fit in                        │        ",
+            "   │       your small                    │        ",
+            "   │       screen                        │        ",
+            "2  │  2    this           678910         │        ",
+            "   │                                     │        ",
+            "   │       is                            │        ",
+            "   │                                     │        ",
+            "   │       an                            │        ",
+            "   │                                     │        ",
+            "   │       even                          │        ",
+            "   │                                     │        ",
+            "   │       longer                        │        ",
+            "   │                                     │        ",
+            "   │       text                          │        ",
+            "   │                                     │        ",
+            "   │       that                          │        ",
+            "   │                                     │        ",
+            "   │       surely                        │        ",
+            "   │                                     │        ",
+            "   │       will                          │        ",
+            "   │                                     │        ",
+            "   │       not                           │        ",
+            "   │                                     │        ",
+            "   │       fit                           │        ",
+            "   │                                     │        ",
+            "   │       in                            │        ",
+            "   │                                     │        ",
+            "   │       your                          │        ",
+            "   │                                     │        ",
+            "   │       small                         │        ",
+            "   │                                     │        ",
+            "   │       screen                        │        ",
+            "3  │  3    normal         123,456,789    │        ",
+            "   │       text now                      │        ",
+            "   │                                     │        ",
+            "───┴─────────────────────────────────────┴────────",
+            "Word wrap enabled                                 ",
+        ];
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        assert_eq!(lines, expected);
+    }
+
+    #[test]
+    fn test_starts_with_newline() {
+        let mut app = AppBuilder::new("tests/data/starts_with_newline.csv")
+            .build()
+            .unwrap();
+        thread::sleep(time::Duration::from_millis(100));
+
+        let backend = TestBackend::new(50, 20);
+        let mut terminal = Terminal::new(backend).unwrap();
+
+        step_and_draw(&mut app, &mut terminal, Control::Nothing);
+        let expected = vec![
+            "──────────────────────────────────────────────────",
+            "      a    b              c                       ",
+            "───┬─────────────────────────────────────┬────────",
+            "1  │  1    this is a …    12345          │        ",
+            "2  │  2    …              678910         │        ",
+            "3  │  3    normal tex…    123,456,789    │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "   │                                     │        ",
+            "───┴─────────────────────────────────────┴────────",
+            "stdin [Row 1/3, Col 1/3]                          ",
+        ];
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        assert_eq!(lines, expected);
+
+        step_and_draw(&mut app, &mut terminal, Control::ToggleLineWrap(true));
+        let expected = vec![
+            "──────────────────────────────────────────────────",
+            "      a    b              c                       ",
+            "───┬─────────────────────────────────────┬────────",
+            "1  │  1    this is a      12345          │        ",
+            "   │       very long                     │        ",
+            "   │       text that                     │        ",
+            "   │       surely                        │        ",
+            "   │       will not                      │        ",
+            "   │       fit in                        │        ",
+            "   │       your small                    │        ",
+            "   │       screen                        │        ",
+            "2  │  2                   678910         │        ",
+            "   │       starts                        │        ",
+            "   │       with new                      │        ",
+            "   │       line                          │        ",
+            "3  │  3    normal         123,456,789    │        ",
+            "   │       text now                      │        ",
+            "   │                                     │        ",
+            "───┴─────────────────────────────────────┴────────",
+            "Word wrap enabled                                 ",
+        ];
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        assert_eq!(lines, expected);
     }
 
     #[test]
@@ -1384,6 +1706,10 @@ mod tests {
         step_and_draw(&mut app, &mut terminal, Control::ScrollRight);
         step_and_draw(&mut app, &mut terminal, Control::ToggleSort);
 
+        // Toggle back to row selection mode before filtering
+        step_and_draw(&mut app, &mut terminal, Control::ToggleSelectionType);
+        step_and_draw(&mut app, &mut terminal, Control::ToggleSelectionType);
+
         thread::sleep(time::Duration::from_millis(200));
         step_and_draw(&mut app, &mut terminal, Control::Nothing);
 
@@ -1409,7 +1735,7 @@ mod tests {
             "88  │  34      25            11      Santa Barbara     │                        ",
             "86  │  38      26            23      Santa Rosa        │                        ",
             "────┴──────────────────────────────────────────────────┴────────────────────────",
-            "stdin [Row 94/128, Col 1/4] [Filter \"San\": -/11] [Filter \"Lat|City\": 4/10 cols] ",
+            "stdin [Row 94/128, Col 1/4] [Filter \"San\": 1/11] [Filter \"Lat|City\": 4/10 cols] ",
         ];
         assert_eq!(lines, expected);
     }
@@ -1504,4 +1830,74 @@ mod tests {
         let lines = to_lines(&actual_buffer);
         assert_eq!(lines, expected);
     }
+
+    #[test]
+    fn test_filter_rows_for_specific_column() {
+        let mut app = AppBuilder::new("tests/data/cities.csv").build().unwrap();
+        thread::sleep(time::Duration::from_millis(100));
+
+        let backend = TestBackend::new(80, 10);
+        let mut terminal = Terminal::new(backend).unwrap();
+
+        step_and_draw(&mut app, &mut terminal, Control::ToggleSelectionType);
+        step_and_draw(&mut app, &mut terminal, Control::ScrollRight);
+        step_and_draw(&mut app, &mut terminal, Control::Filter("^1".into()));
+
+        thread::sleep(time::Duration::from_millis(200));
+        step_and_draw(&mut app, &mut terminal, Control::Nothing);
+
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        let expected = vec![
+            "────────────────────────────────────────────────────────────────────────────────",
+            "       LatD    LatM    LatS    NS    LonD    LonM    LonS    EW    City         ",
+            "────┬───────────────────────────────────────────────────────────────────────────",
+            "4   │  42      16      12      N     71      48      0       W     Worcester    ",
+            "8   │  39      11      23      N     78      9       36      W     Winchest…    ",
+            "9   │  34      14      24      N     77      55      11      W     Wilmingt…    ",
+            "12  │  41      15      0       N     77      0       0       W     Williams…    ",
+            "20  │  31      13      11      N     82      20      59      W     Waycross     ",
+            "────┴───────────────────────────────────────────────────────────────────────────",
+            "stdin [Row 4/128, Col 1/10] [Filter \"^1\" in LatM: -/19]                         ",
+        ];
+        assert_eq!(lines, expected);
+    }
+
+    #[test]
+    fn test_filter_like_cell() {
+        let mut app = AppBuilder::new("tests/data/cities.csv").build().unwrap();
+        thread::sleep(time::Duration::from_millis(100));
+
+        let backend = TestBackend::new(80, 10);
+        let mut terminal = Terminal::new(backend).unwrap();
+
+        // Enter cell selection mode
+        step_and_draw(&mut app, &mut terminal, Control::ToggleSelectionType);
+        step_and_draw(&mut app, &mut terminal, Control::ToggleSelectionType);
+
+        // Select the State column
+        for _ in 0..10 {
+            step_and_draw(&mut app, &mut terminal, Control::ScrollRight);
+        }
+        step_and_draw(&mut app, &mut terminal, Control::FilterLikeCell);
+
+        thread::sleep(time::Duration::from_millis(200));
+        step_and_draw(&mut app, &mut terminal, Control::Nothing);
+
+        let actual_buffer = terminal.backend().buffer().clone();
+        let lines = to_lines(&actual_buffer);
+        let expected = vec![
+            "────────────────────────────────────────────────────────────────────────────────",
+            "       LatS    NS    LonD    LonM    LonS    EW    City            State        ",
+            "────┬───────────────────────────────────────────────────────────────────────┬───",
+            "1   │  59      N     80      39      0       W     Youngstown      OH       │   ",
+            "50  │  0       N     83      32      24      W     Toledo          OH       │   ",
+            "62  │  36      N     80      37      12      W     Steubenville    OH       │   ",
+            "65  │  11      N     83      48      35      W     Springfield     OH       │   ",
+            "92  │  0       N     82      42      35      W     Sandusky        OH       │   ",
+            "────┴───────────────────────────────────────────────────────────────────────┴───",
+            "stdin [Row 1/128, Col 3/10] [Filter \"OH\" in State: 1/6]                         ",
+        ];
+        assert_eq!(lines, expected);
+    }
 }
