@@ -119,30 +119,43 @@ impl<'a> CsvTable<'a> {
 
     fn get_row_heights(
         &self,
+        area_height: u16,
         rows: &[Row],
         column_widths: &[u16],
         enable_line_wrap: bool,
+        is_word_wrap: bool,
     ) -> Vec<u16> {
         if !enable_line_wrap {
             return rows.iter().map(|_| 1).collect();
         }
+        let mut total_height = 0;
         let mut row_heights = Vec::new();
         for (i, row) in rows.iter().enumerate() {
+            if total_height >= area_height {
+                // Exit early if we've already filled the available height. Important since
+                // LineWrapper at its current state is not particularly efficient...
+                row_heights.push(1);
+                continue;
+            }
             for (j, content) in row.fields.iter().enumerate() {
-                let mut num_lines = 0;
-                for parts in content.split('\n') {
-                    let parts_length = parts.chars().count();
-                    num_lines += max(
-                        1,
-                        match column_widths.get(j) {
-                            Some(w) => {
-                                let usable_width = (*w).saturating_sub(NUM_SPACES_BETWEEN_COLUMNS);
-                                (parts_length as f32 / usable_width as f32).ceil() as u16
+                let num_lines = match column_widths.get(j) {
+                    Some(w) => {
+                        let usable_width = (*w).saturating_sub(NUM_SPACES_BETWEEN_COLUMNS);
+                        let spans = [Span::styled(content.as_str(), Style::default())];
+                        let mut line_wrapper =
+                            wrap::LineWrapper::new(&spans, usable_width as usize, is_word_wrap);
+                        let mut num_lines = 0;
+                        loop {
+                            line_wrapper.next();
+                            num_lines += 1;
+                            if line_wrapper.finished() {
+                                break;
                             }
-                            None => 1,
-                        },
-                    );
-                }
+                        }
+                        num_lines
+                    }
+                    None => 1,
+                };
                 if let Some(height) = row_heights.get_mut(i) {
                     if *height < num_lines {
                         *height = num_lines;
@@ -151,6 +164,7 @@ impl<'a> CsvTable<'a> {
                     row_heights.push(num_lines);
                 }
             }
+            total_height += row_heights[i];
         }
         row_heights
     }
@@ -337,12 +351,19 @@ impl<'a> CsvTable<'a> {
                 style: filler_style,
                 short_padding,
             };
+
+            let should_highlight_cell = |active: &FinderActiveState, content: &str| {
+                if let Some((target_column_index, _)) = active.column_index {
+                    if target_column_index != col_index {
+                        return false;
+                    }
+                }
+                active.target.is_match(content) && !matches!(row_type, RowType::Header)
+            };
             match &state.finder_state {
                 // TODO: seems like doing a bit too much of heavy lifting of
                 // checking for matches (finder's work)
-                FinderState::FinderActive(active)
-                    if active.target.is_match(hname) && !matches!(row_type, RowType::Header) =>
-                {
+                FinderState::FinderActive(active) if should_highlight_cell(active, hname) => {
                     let mut highlight_style = filler_style.style.fg(Color::Rgb(200, 0, 0));
                     if let Some(hl) = &active.found_record {
                         if let Some(row_index) = row_index {
@@ -368,6 +389,7 @@ impl<'a> CsvTable<'a> {
                         effective_width,
                         row_height,
                         filler_style,
+                        state.is_word_wrap,
                     );
                 }
                 _ => {
@@ -380,6 +402,7 @@ impl<'a> CsvTable<'a> {
                         effective_width,
                         row_height,
                         filler_style,
+                        state.is_word_wrap,
                     );
                 }
             };
@@ -462,6 +485,7 @@ impl<'a> CsvTable<'a> {
         width: u16,
         height: u16,
         filler_style: FillerStyle,
+        is_word_wrap: bool,
     ) {
         const SUFFIX: &str = "…";
         const SUFFIX_LEN: u16 = 1;
@@ -475,7 +499,8 @@ impl<'a> CsvTable<'a> {
             NUM_SPACES_BETWEEN_COLUMNS
         } as usize;
 
-        let mut line_wrapper = wrap::LineWrapper::new(spans, effective_width as usize);
+        let mut line_wrapper =
+            wrap::LineWrapper::new(spans, effective_width as usize, is_word_wrap);
 
         for offset in 0..height {
             if let Some(mut line) = line_wrapper.next() {
@@ -636,13 +661,22 @@ impl<'a> CsvTable<'a> {
         buf.set_span(area.x, area.bottom().saturating_sub(1), &span, area.width);
     }
 
-    fn get_view_layout(&self, area: Rect, state: &CsvTableState) -> ViewLayout {
+    fn get_view_layout(&self, area: Rect, state: &mut CsvTableState) -> ViewLayout {
         let column_widths = self.get_column_widths(
             area.width,
             &state.column_width_overrides,
             &state.sorter_state,
         );
-        let row_heights = self.get_row_heights(self.rows, &column_widths, state.enable_line_wrap);
+        let _tic = std::time::Instant::now();
+        let row_heights = self.get_row_heights(
+            area.height,
+            self.rows,
+            &column_widths,
+            state.enable_line_wrap,
+            state.is_word_wrap,
+        );
+        // state.debug = format!("get_row_heights elapsed: {:?}", _tic.elapsed());
+        // state.debug = format!("row_heights: {:?}, area_height: {:?}", row_heights, area.height);
         ViewLayout {
             column_widths,
             row_heights,
@@ -794,6 +828,7 @@ pub struct FinderActiveState {
     total_found: u64,
     cursor_index: Option<u64>,
     target: Regex,
+    column_index: Option<(usize, String)>,
     found_record: Option<find::FoundRecord>,
     selected_offset: Option<u64>,
     is_filter: bool,
@@ -806,6 +841,9 @@ impl FinderActiveState {
             total_found: finder.count() as u64,
             cursor_index: finder.cursor().map(|x| x as u64),
             target: finder.target(),
+            column_index: finder
+                .column_index()
+                .map(|i| (i, rows_view.get_column_name_from_global_index(i))),
             found_record: finder.current(),
             selected_offset: rows_view.selected_offset(),
             is_filter: rows_view.is_filter(),
@@ -842,7 +880,12 @@ impl FinderActiveState {
             line = format!("{cursor_str}/{}{plus_marker}", self.total_found);
         }
         let action = if self.is_filter { "Filter" } else { "Find" };
-        format!("[{action} \"{}\": {line}]", self.target)
+        let target_column = self
+            .column_index
+            .as_ref()
+            .map(|(_, name)| format!(" in {}", name))
+            .unwrap_or_default();
+        format!("[{action} \"{}\"{target_column}: {line}]", self.target)
     }
 }
 
@@ -985,6 +1028,7 @@ pub struct CsvTableState {
     pub no_headers: bool,
     pub view_layout: Option<ViewLayout>,
     pub enable_line_wrap: bool,
+    pub is_word_wrap: bool,
     pub column_width_overrides: ColumnWidthOverrides,
     pub debug: String,
 }
@@ -1019,6 +1063,7 @@ impl CsvTableState {
             no_headers,
             view_layout: None,
             enable_line_wrap: false,
+            is_word_wrap: false,
             column_width_overrides: ColumnWidthOverrides::new(),
             debug: "".into(),
         }
